"""This module provides various logging utilities and initializers for the liblaf library.

The module includes functions and classes for setting up and configuring logging using
Loguru, Rich, and IceCream libraries. It also provides utilities for retrieving caller
locations and fully qualified names of objects.
"""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach_stub(__name__, __file__)



import logging
from collections.abc import Sequence

import loguru

from . import init_icecream, init_loguru, init_rich


def init_logging(
    level: int | str = logging.NOTSET,
    *,
    handlers: Sequence["loguru.HandlerConfig"] | None = None,
    levels: Sequence["loguru.LevelConfig"] | None = None,
    traceback_show_locals: bool = True,
) -> None:
    """Initializes the logging configuration for the application.

    This function sets up logging using Rich, Loguru, and IceCream libraries.

    Args:
        level: The logging level.
        handlers: A sequence of Loguru handler configurations.
        levels: A sequence of Loguru level configurations.
        traceback_show_locals: Whether to show local variables in tracebacks.
    """
    init_rich(show_locals=traceback_show_locals)
    init_loguru(level=level, handlers=handlers, levels=levels)
    init_icecream()



from collections.abc import Generator, Iterable

from rich.progress import Progress

from liblaf import grapes

from . import progress


def track[T](
    sequence: Iterable[T],
    *,
    description: str | bool | None = True,
    record_log_level: int | str | None = "DEBUG",
    summary_log_level: int | str | None = "INFO",
    timer: bool = True,
    total: float | None = None,
) -> Generator[T]:
    """Tracks the progress of iterating over a sequence with optional logging and timing.

    Args:
        sequence: The sequence to iterate over.
        description: Description for the progress bar. If `True`, uses caller location.
        record_log_level: Log level for recording progress.
        summary_log_level: Log level for reporting progress.
        timer: Whether to use a timer for tracking.
        total: Total number of items in the sequence.

    Yields:
        Yields items from the sequence.
    """
    if description is True:
        description = grapes.caller_location(2)
    description = description or ""
    prog: Progress = progress()
    if timer:
        t = grapes.timer(
            label=description,
            log_summary_at_exit=False,
            record_log_level=record_log_level,
            summary_log_level=summary_log_level,
        )
        with prog:
            yield from t.track(
                prog.track(sequence, total=total, description=description)
            )
            t.log_summary()
    else:
        with prog:
            yield from prog.track(sequence, description=description)



import functools
from collections.abc import Callable

from liblaf import grapes

from . import TimerAttrs, TimerTrait


class TimedFunction[**P, T](TimerTrait):
    fn: Callable[P, T]
    __timer_attrs: TimerAttrs

    @property
    def _timer_attrs(self) -> TimerAttrs:
        return self.__timer_attrs

    def __init__(self, fn: Callable[P, T], timer: TimerTrait) -> None:
        self.__timer_attrs = timer._timer_attrs  # noqa: SLF001
        self.fn = fn
        self.label = self.label or grapes.full_qual_name(fn) or "Function"
        functools.update_wrapper(self, fn)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:  # pyright: ignore[reportIncompatibleMethodOverride]
        self.start()
        result: T = self.fn(*args, **kwargs)
        self.end(depth=3)
        return result



import os
from collections.abc import Callable
from pathlib import Path
from typing import Any

from liblaf import grapes

READERS: dict[str, Callable] = {}
"""A registry mapping file extensions to their respective deserialization functions.

Supported extensions and their corresponding deserialization functions:

- ".json": grapes.load_json
- ".toml": grapes.load_toml (if the "tomlkit" module is available)
- ".yaml": grapes.load_yaml (if the "ruamel.yaml" module is available)
- ".yml": grapes.load_yaml (if the "ruamel.yaml" module is available)
"""

WRITERS: dict[str, Callable] = {}
"""A registry mapping file extensions to their respective serialization functions.

Supported extensions and their corresponding serialization functions:

- ".json": grapes.save_json
- ".toml": grapes.save_toml (if the "tomlkit" module is available)
- ".yaml": grapes.save_yaml (if the "ruamel.yaml" module is available)
- ".yml": grapes.save_yaml (if the "ruamel.yaml" module is available)
"""


READERS[".json"] = grapes.load_json
WRITERS[".json"] = grapes.save_json
if grapes.has_module("tomlkit"):
    READERS[".toml"] = grapes.load_toml
    WRITERS[".toml"] = grapes.save_toml
if grapes.has_module("ruamel.yaml"):
    READERS[".yaml"] = grapes.load_yaml
    READERS[".yml"] = grapes.load_yaml
    WRITERS[".yaml"] = grapes.save_yaml
    WRITERS[".yml"] = grapes.save_yaml


def serialize(
    fpath: str | os.PathLike[str], data: Any, *, ext: str | None = None
) -> None:
    """Serialize data to a file.

    Args:
        fpath: The file path where the data will be serialized.
        data: The data to be serialized.
        ext: The file extension to determine the writer to use. If `None`, the extension is derived from the file path.

    Raises:
        ValueError: If the file extension is not supported.
    """
    fpath: Path = Path(fpath)
    if ext is None:
        ext = fpath.suffix
    if ext not in WRITERS:
        msg: str = f"Unsupported file extension: {ext}"
        raise ValueError(msg)
    writer = WRITERS[ext]
    fpath.parent.mkdir(parents=True, exist_ok=True)
    writer(fpath, data)


def deserialize(fpath: str | os.PathLike[str], *, ext: str | None = None) -> Any:
    """Deserialize the contents of a file.

    Args:
        fpath: The path to the file to be deserialized.
        ext: The file extension. If not provided, it will be inferred from the file path.

    Returns:
        The deserialized content of the file.

    Raises:
        ValueError: If the file extension is not supported.
    """
    fpath: Path = Path(fpath)
    if ext is None:
        ext = fpath.suffix
    if ext not in READERS:
        msg: str = f"Unsupported file extension: {ext}"
        raise ValueError(msg)
    return READERS[ext](fpath)



import abc
import collections
import functools
import statistics
import textwrap
from collections.abc import Iterator, Mapping, Sequence
from typing import overload

import attrs
from loguru import logger

from liblaf import grapes


@attrs.define
class TimerRecordsAttrs:
    label: str | None = None
    log_summary_at_exit: bool = False
    record_log_level: int | str | None = "DEBUG"
    records: dict[str, list[float]] = attrs.field(
        factory=lambda: collections.defaultdict(list)
    )
    summary_log_level: int | str | None = "INFO"


class TimerRecordsTrait(abc.ABC):
    # region attrs

    @property
    @abc.abstractmethod
    def _timer_records_attrs(self) -> TimerRecordsAttrs: ...

    @property
    def label(self) -> str | None:
        return self._timer_records_attrs.label

    @label.setter
    def label(self, value: str | None) -> None:
        self._timer_records_attrs.label = value

    @property
    def log_summary_at_exit(self) -> bool:
        return self._timer_records_attrs.log_summary_at_exit

    @property
    def record_log_level(self) -> int | str | None:
        return self._timer_records_attrs.record_log_level

    @property
    def summary_log_level(self) -> int | str | None:
        return self._timer_records_attrs.summary_log_level

    @property
    def _records(self) -> dict[str, list[float]]:
        return self._timer_records_attrs.records

    # endregion attrs

    @overload
    def __getitem__(self, index: int) -> Mapping[str, float]: ...
    @overload
    def __getitem__(self, index: str) -> Sequence[float]: ...
    def __getitem__(self, index: int | str) -> Mapping[str, float] | Sequence[float]:
        if isinstance(index, int):
            return self.row(index)
        return self.column(index)

    def __len__(self) -> int:
        return self.count

    @property
    def columns(self) -> Sequence[str]:
        return list(self._records.keys())

    @property
    def count(self) -> int:
        return self.n_rows

    @functools.cached_property
    def default_key(self) -> str:
        return next(iter(self._records))

    @property
    def n_columns(self) -> int:
        return len(self._records)

    @property
    def n_rows(self) -> int:
        return len(self.column())

    def append(
        self, seconds: Mapping[str, float] = {}, nanoseconds: Mapping[str, float] = {}
    ) -> None:
        for key, value in seconds.items():
            self._records[key].append(value)
        for key, value in nanoseconds.items():
            self._records[key].append(value * 1e-9)

    def column(self, key: str | None = None) -> Sequence[float]:
        return self._records[key or self.default_key]

    def human_record(self, index: int = -1, label: str | None = None) -> str:
        label: str = self.label or "Timer"
        text: str = f"{label} > "
        for key, value in self.row(index).items():
            human_duration: str = grapes.human_duration(value)
            text += f"{key}: {human_duration}, "
        text = text.strip(", ")
        return text

    def human_summary(self, label: str | None = None) -> str:
        label: str = label or self.label or "Timer"
        header: str = f"{label} (total: {self.n_rows})"
        if self.n_rows == 0:
            return header
        body: str = ""
        for k in self.columns:
            body += f"{k} > "
            human_mean: str = grapes.human_duration_series(self.column(k))
            human_best: str = grapes.human_duration(self.min(k))
            body += f"mean: {human_mean}, best: {human_best}\n"
        body = body.strip()
        summary: str = header + "\n" + textwrap.indent(body, "  ")
        return summary

    def iter_columns(self) -> Iterator[tuple[str, Sequence[float]]]:
        yield from self._records.items()

    def iter_rows(self) -> Iterator[Mapping[str, float]]:
        for index in range(self.n_rows):
            yield self.row(index)

    def log_record(
        self,
        index: int = -1,
        label: str | None = None,
        depth: int = 1,
        level: int | str | None = None,
    ) -> None:
        level = level or self.record_log_level
        if level is None:
            return
        logger.opt(depth=depth).log(level, self.human_record(index=index, label=label))

    def log_summary(
        self, label: str | None = None, depth: int = 1, level: int | str | None = None
    ) -> None:
        level = level or self.summary_log_level
        if level is None:
            return
        logger.opt(depth=depth).log(level, self.human_summary(label=label))

    def row(self, index: int = -1) -> Mapping[str, float]:
        return {key: values[index] for key, values in self._records.items()}

    # region statistics

    def max(self, key: str | None = None) -> float:
        return max(self.column(key))

    def mean(self, key: str | None = None) -> float:
        return statistics.mean(self.column(key))

    def min(self, key: str | None = None) -> float:
        return min(self.column(key))

    def std(self, key: str | None = None) -> float:
        return statistics.stdev(self.column(key))

    # endregion statistics



from collections.abc import Callable, Iterable

from . import TimedFunction, TimedIterable, TimerAttrs, TimerTrait, register_timer


class timer(TimerTrait):  # noqa: N801
    __timer_attrs: TimerAttrs

    @property
    def _timer_attrs(self) -> TimerAttrs:
        return self.__timer_attrs

    def __init__(
        self,
        label: str | None = None,
        *,
        counters: Iterable[str] = ["perf", "process"],
        log_summary_at_exit: bool = False,
        record_log_level: int | str | None = "DEBUG",
        summary_log_level: int | str | None = "INFO",
    ) -> None:
        self.__timer_attrs = TimerAttrs(
            label=label,
            log_summary_at_exit=log_summary_at_exit,
            record_log_level=record_log_level,
            summary_log_level=summary_log_level,
        )
        for counter in counters:
            self._records[counter] = []
        if self.log_summary_at_exit:
            register_timer(self)

    def __call__[**P, T](self, fn: Callable[P, T]) -> TimedFunction[P, T]:
        return TimedFunction(fn, self)

    def track[T](self, iterable: Iterable[T]) -> TimedIterable[T]:
        return TimedIterable(iterable, self)



import math
import statistics
from collections.abc import Sequence

UNITS: dict[str, float] = {
    "ns": 1e-9,
    "us": 1e-6,
    "ms": 1e-3,
    "s": 1,
    "m": 60,
    "h": 3600,
    "d": 86400,
    "w": 604800,
    "y": 31536000,
}


def get_unit_seconds(unit: str) -> float:
    """Convert a time unit to its equivalent in seconds.

    Args:
        unit : The time unit to convert (e.g., 'm', 'h').

    Returns:
        The equivalent time in seconds.

    Raises:
        KeyError: If the provided unit is not found in the `UNITS` dictionary.
    """
    return UNITS[unit.lower()]


def human_duration_unit_precision(seconds: float) -> tuple[str, int]:  # noqa: C901, PLR0911, PLR0912
    """Determine the appropriate human-readable duration unit and precision for a given time in seconds.

    Args:
        seconds: The duration in seconds.

    Returns:
        A tuple containing the unit of time as a string and the precision as an integer. The unit can be "ns" (nanoseconds), "us" (microseconds), "ms" (milliseconds), "s" (seconds), "m" (minutes), or "h" (hours). The precision indicates the number of decimal places to display for the given unit.
    """
    if seconds <= 0:
        return "s", 0
    if seconds < 1e-09:
        return "ns", 3  # .999 ns
    if seconds < 1e-08:
        return "ns", 2  # 9.99 ns
    if seconds < 1e-07:
        return "ns", 1  # 99.9 ns
    if seconds < 1e-06:
        return "ns", 0  # 999. ns
    if seconds < 1e-05:
        return "us", 2  # 9.99 us
    if seconds < 1e-04:
        return "us", 1  # 99.9 us
    if seconds < 1e-03:
        return "us", 0  # 999. us
    if seconds < 1e-02:
        return "ms", 2  # 9.99 ms
    if seconds < 1e-01:
        return "ms", 1  # 99.9 ms
    if seconds < 1:
        return "ms", 0  # 999. ms
    if seconds < 10:
        return "s", 2  # 9.99 s
    if seconds < 60:
        return "s", 1  # 59.9 s
    if seconds < 3600:
        return "m", 0  # 59:59
    if seconds < 86400:
        return "h", 0  # 23:59:59
    return "h", 0


def human_duration(
    seconds: float, unit: str | None = None, precision: int | None = None
) -> str:
    """Convert a duration in seconds to a human-readable string format.

    Args:
        seconds: The duration in seconds to be converted.
        unit: The unit of time to use for the output. Can be one of {"ns", "us", "ms", "s", "m"}. If `None`, the unit will be determined automatically based on the duration.
        precision: The number of decimal places to include in the output. If `None`, the precision will be determined automatically based on the duration.

    Returns:
        The human-readable duration string.

    Notes:
        - For units "ns", "us", "ms", and "s", the output will be a floating-point number followed by the unit.
        - For unit "m", the output will be in the format "MM:SS".
        - For durations longer than an hour, the output will be in the format "HH:MM:SS".
        - The function currently does not handle durations longer than a day.
    """
    if (unit is None) or (precision is None):
        unit, precision = human_duration_unit_precision(seconds)
    if unit in {"ns", "us", "ms", "s"}:
        unit_seconds: float = get_unit_seconds(unit)
        value: float = seconds / unit_seconds
        human: str = f"{value:.{precision}f}".lstrip("0")
        if precision == 0:
            human += "."
        if unit == "us":
            unit = "µs"
        human += f" {unit}"
        return human
    if unit == "m":
        minutes: int = int(seconds // 60)
        seconds %= 60
        return f"{minutes}:{seconds:02.0f}"
    hours: int = int(seconds // 3600)
    seconds %= 3600
    minutes: int = int(seconds // 60)
    seconds %= 60
    return f"{hours}:{minutes:02.0f}:{seconds:02.0f}"
    # TODO: handle longer durations


def human_duration_with_variance(mean: float, std: float) -> str:
    """Returns a human-readable string representing a duration with variance.

    Args:
        mean: The mean duration.
        std: The standard deviation of the duration.

    Returns:
        A string representing the mean duration with its variance in the format "mean ± std". If the standard deviation is not finite, only the mean duration is returned.
    """
    if not math.isfinite(std):
        return human_duration(mean)
    return human_duration(mean) + " ± " + human_duration(std)


def human_duration_series(series: Sequence[float]) -> str:
    """Convert a series of durations into a human-readable string.

    Args:
        series: An array-like object containing duration values.

    Returns:
        A human-readable string representing the duration. If the series contains only one element, it returns the human-readable format of that single duration. If the series contains more than one element, it returns the mean duration with its variance in a human-readable format.
    """
    if len(series) <= 1:
        return human_duration(series[0])
    return human_duration_with_variance(
        statistics.mean(series), statistics.stdev(series)
    )



import json
import os
from pathlib import Path
from typing import Any


def load_json(fpath: str | os.PathLike[str]) -> Any:
    """Load and parse a JSON file from the given file path.

    Args:
        fpath: The path to the JSON file.

    Returns:
        Any: The parsed JSON data.

    Raises:
        FileNotFoundError: If the file does not exist.
        json.JSONDecodeError: If the file contains invalid JSON.
    """
    fpath: Path = Path(fpath)
    with fpath.open() as fp:
        return json.load(fp)


def save_json(fpath: str | os.PathLike[str], data: Any) -> None:
    """Save data to a JSON file.

    Args:
        fpath: The file path where the JSON data will be saved.
        data: The data to be serialized to JSON and saved to the file.
    """
    fpath: Path = Path(fpath)
    with fpath.open("w") as fp:
        json.dump(data, fp)



from ._count import human_count
from ._duration import (
    get_unit_seconds,
    human_duration,
    human_duration_series,
    human_duration_unit_precision,
    human_duration_with_variance,
)
from ._throughout import human_throughout

__all__ = [
    "get_unit_seconds",
    "human_count",
    "human_duration",
    "human_duration_series",
    "human_duration_unit_precision",
    "human_duration_with_variance",
    "human_throughout",
]



from ._as_iterable import as_iterable
from ._as_sequence import as_sequence
from ._generator_to_list import generator_to_list
from ._iterable_wrapper import IterableWrapper

__all__ = ["IterableWrapper", "as_iterable", "as_sequence", "generator_to_list"]



from loguru import logger

from liblaf import grapes


def init_icecream() -> None:
    """Initializes the icecream debugging tool if the 'icecream' module is available.

    This function checks if the 'icecream' module is present in the 'grapes' package. If the module is available, it configures the output of the icecream debugger to use the custom logger with a specific log level "ICECREAM".
    """
    if not grapes.has_module("icecream"):
        return
    from icecream import ic

    ic.configureOutput(
        prefix="", outputFunction=lambda s: logger.opt(depth=2).log("ICECREAM", s)
    )



import atexit

from . import TimerRecordsTrait

TIMERS: list[TimerRecordsTrait] = []


def register_timer(timer: TimerRecordsTrait) -> None:
    TIMERS.append(timer)


def log_summary() -> None:
    for timer in TIMERS:
        if timer.log_summary_at_exit and timer.n_rows > 1:
            timer.log_summary()


atexit.register(log_summary)



from collections.abc import Collection, Iterable, Iterator


class IterableWrapper[T](Collection[T]):
    iterable: Iterable[T]

    def __init__(self, iterable: Iterable[T], /) -> None:
        self.iterable = iterable

    def __contains__(self, x: object, /) -> bool:
        return x in self.iterable  # pyright: ignore[reportOperatorIssue]

    def __iter__(self) -> Iterator[T]:
        return iter(self.iterable)

    def __len__(self) -> int:
        return len(self.iterable)  # pyright: ignore[reportArgumentType]



from ._caller import caller_location
from ._icecream import init_icecream
from ._init import init_logging
from ._loguru import (
    InterceptHandler,
    add_level,
    init_loguru,
    setup_loguru_logging_intercept,
)
from ._name import full_qual_name
from ._once import (
    critical_once,
    debug_once,
    error_once,
    exception_once,
    info_once,
    log_once,
    success_once,
    trace_once,
    warning_once,
)
from ._rich import init_rich, logging_console, logging_theme

__all__ = [
    "InterceptHandler",
    "add_level",
    "caller_location",
    "critical_once",
    "debug_once",
    "error_once",
    "exception_once",
    "full_qual_name",
    "info_once",
    "init_icecream",
    "init_logging",
    "init_loguru",
    "init_rich",
    "log_once",
    "logging_console",
    "logging_theme",
    "setup_loguru_logging_intercept",
    "success_once",
    "trace_once",
    "warning_once",
]



import functools

from environs import Env


@functools.lru_cache
def init_env(prefix: str | None = "LIBLAF_") -> Env:
    """Initialize and return an environment configuration.

    This function creates an instance of the Env class with the specified prefix, reads the environment variables, and returns the configured environment.

    Args:
        prefix: The prefix to use for environment variables.

    Returns:
        The initialized environment configuration.
    """
    env = Env(prefix=prefix)
    env.read_env()
    return env



def full_qual_name(obj: object) -> str:
    """Returns the fully qualified name of an object, including its module and class.

    Args:
        obj: The object for which to get the fully qualified name.

    Returns:
        The fully qualified name of the object, including its module and class. If the object is callable, "()" is appended to the name.
    """
    name: str = obj.__module__ + "." + obj.__qualname__
    if callable(obj):
        name += "()"
    return name



from pathlib import Path
from types import FrameType

from loguru._get_frame import get_frame


def caller_location(depth: int = 1, *, markup: bool = True) -> str:
    """Returns the file name and line number of the caller's location in the code.

    Args:
        depth: The stack depth to inspect.
        markup: If `True`, returns the file name and line number with markup for links.

    Returns:
        The file name and line number of the caller's location. If the frame cannot be retrieved, returns "<unknown>".
    """
    frame: FrameType | None = get_frame(depth)
    if not frame:
        return "<unknown>"
    filepath: Path = Path(frame.f_code.co_filename)
    if markup:
        filename: str = f"[link={filepath.as_uri()}]{filepath.name}[/link]"
        lineno: str = (
            f"[link={filepath.as_uri()}#{frame.f_lineno}]{frame.f_lineno}[/link]"
        )
    else:
        filename = filepath.name
        lineno = str(frame.f_lineno)
    return f"{filename}:{lineno}"



from ._json import load_json, save_json
from ._pydantic import load_pydantic, save_pydantic
from ._serde import READERS, WRITERS, deserialize, serialize
from ._toml import load_toml, save_toml
from ._yaml import load_yaml, save_yaml

__all__ = [
    "READERS",
    "WRITERS",
    "deserialize",
    "load_json",
    "load_pydantic",
    "load_toml",
    "load_yaml",
    "save_json",
    "save_pydantic",
    "save_toml",
    "save_yaml",
    "serialize",
]



import abc
import contextlib
import types
from typing import Self

import attrs

from . import TimerRecordsAttrs, TimerRecordsTrait, get_time


@attrs.define
class TimerAttrs(TimerRecordsAttrs):
    end: dict[str, float] = attrs.field(factory=dict)
    start: dict[str, float] = attrs.field(factory=dict)


class TimerTrait(contextlib.AbstractContextManager, TimerRecordsTrait):
    # region attrs

    @property
    @abc.abstractmethod
    def _timer_attrs(self) -> TimerAttrs: ...

    @property
    def _end(self) -> dict[str, float]:
        return self._timer_attrs.end

    @property
    def _start(self) -> dict[str, float]:
        return self._timer_attrs.start

    @property
    def _timer_records_attrs(self) -> TimerRecordsAttrs:
        return self._timer_attrs

    # endregion attrs

    def __enter__(self) -> Self:
        self.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        self.end(depth=3)

    def start(self) -> None:
        for counter in self.columns:
            self._start[counter] = get_time(counter)

    def end(self, depth: int = 2) -> None:
        for counter in self.columns:
            self._end[counter] = get_time(counter)
        self.append(
            {
                counter: self._end[counter] - self._start[counter]
                for counter in self.columns
            }
        )
        self.log_record(depth=depth)



import os
from pathlib import Path
from typing import Any

from liblaf import grapes

with grapes.optional_imports(extra="serde"):
    from ruamel.yaml import YAML


def load_yaml(fpath: str | os.PathLike[str]) -> Any:
    fpath: Path = Path(fpath)
    yaml = YAML()
    with fpath.open() as fp:
        return yaml.load(fp)


def save_yaml(fpath: str | os.PathLike[str], data: Any) -> None:
    fpath: Path = Path(fpath)
    yaml = YAML()
    with fpath.open("w") as fp:
        yaml.dump(data, fp)



import os
from pathlib import Path
from typing import Any

from liblaf import grapes

with grapes.optional_imports(extra="toml"):
    import tomlkit


def load_toml(fpath: str | os.PathLike[str]) -> tomlkit.TOMLDocument:
    """Load a TOML file and return its contents as a TOMLDocument.

    Args:
        fpath: The file path to the TOML file.

    Returns:
        The contents of the TOML file as a TOMLDocument.
    """
    fpath: Path = Path(fpath)
    with fpath.open() as fp:
        return tomlkit.load(fp)


def save_toml(
    fpath: str | os.PathLike[str], data: Any, *, sort_keys: bool = False
) -> None:
    """Save data to a TOML file.

    Args:
        fpath: The file path where the TOML data will be saved.
        data: The data to be serialized and saved in TOML format.
        sort_keys: Whether to sort the keys in the output.
    """
    fpath: Path = Path(fpath)
    with fpath.open("w") as fp:
        tomlkit.dump(data, fp, sort_keys=sort_keys)



from rich.console import Console, RenderableType
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    Task,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.table import Column
from rich.text import Text

from liblaf import grapes


class RateColumn(ProgressColumn):
    """RateColumn is a subclass of ProgressColumn that represents the rate of progress for a given task."""

    unit: str = "it"
    """The unit of measurement for the progress bar."""

    def __init__(self, unit: str = "it", table_column: Column | None = None) -> None:
        """.

        Args:
            unit: The unit of measurement for the progress bar.
            table_column: The table column associated with the progress bar.
        """
        super().__init__(table_column)
        self.unit = unit

    def render(self, task: Task) -> RenderableType:
        """Render the progress speed of a given task.

        Args:
            task: The task for which the speed is to be rendered.

        Returns:
            A text object representing the speed of the task.
        """
        if not task.speed:
            return Text(f"?{self.unit}/s", style="progress.data.speed")
        human: str = grapes.human_throughout(task.speed, self.unit)
        return Text(human, style="progress.data.speed")


def progress(
    *columns: str | ProgressColumn, console: Console | None = None
) -> Progress:
    """Create and return a Progress instance with specified columns and console.

    If no columns are provided, a default set of columns will be used:

    - `SpinnerColumn`
    - `TextColumn` with task description
    - `BarColumn`
    - `TaskProgressColumn` with speed display
    - `MofNCompleteColumn`
    - `TimeElapsedColumn`
    - `TimeRemainingColumn`
    - `RateColumn`

    Args:
        *columns: Variable length argument list of columns to include in the progress display.
        console: The console to use for rendering the progress. Defaults to `None`, in which case the logging console is used.

    Returns:
        An instance of the Progress class configured with the specified columns and console.
    """
    if not columns:
        columns: list[ProgressColumn] = [SpinnerColumn()]
        columns.append(TextColumn("[progress.description]{task.description}"))
        columns += [
            BarColumn(),
            TaskProgressColumn(show_speed=True),
            MofNCompleteColumn(),
            "[",
            TimeElapsedColumn(),
            "<",
            TimeRemainingColumn(),
            ",",
            RateColumn(),
            "]",
        ]
    console = console or grapes.logging_console()
    progress = Progress(*columns, console=console)
    return progress



import functools

import rich.pretty
import rich.traceback
from rich.console import Console
from rich.style import Style
from rich.theme import Theme


@functools.cache
def logging_theme() -> Theme:
    """Returns a Theme object that defines the styles for different logging levels.

    The styles are inspired by the loguru library and include the following levels:

    - notset: Dimmed style
    - trace: Cyan color, bold
    - debug: Blue color, bold
    - icecream: Magenta color, bold
    - info: Bold
    - success: Green color, bold
    - warning: Yellow color, bold
    - error: Red color, bold
    - critical: Red color, bold, reversed

    Returns:
        A Theme object with the specified styles for logging levels.
    """
    # [loguru/loguru/_defaults.py at c490ce0534c6e176306f339a92c221dc6f41a6a7 · Delgan/loguru](https://github.com/Delgan/loguru/blob/c490ce0534c6e176306f339a92c221dc6f41a6a7/loguru/_defaults.py)
    return Theme(
        {
            "logging.level.notset": Style(dim=True),
            "logging.level.trace": Style(color="cyan", bold=True),
            "logging.level.debug": Style(color="blue", bold=True),
            "logging.level.icecream": Style(color="magenta", bold=True),
            "logging.level.info": Style(bold=True),
            "logging.level.success": Style(color="green", bold=True),
            "logging.level.warning": Style(color="yellow", bold=True),
            "logging.level.error": Style(color="red", bold=True),
            "logging.level.critical": Style(color="red", bold=True, reverse=True),
        }
    )


@functools.cache
def logging_console() -> Console:
    """Create and return a rich Console object configured for logging.

    The console is set to use a custom logging theme and output to stderr.

    Returns:
        A rich Console object configured for logging.
    """
    return rich.console.Console(theme=logging_theme(), stderr=True)


def init_rich(*, show_locals: bool = True) -> None:
    """Initialize rich logging for pretty printing and tracebacks.

    This function sets up rich's pretty printing and traceback handling
    for the logging console.

    Args:
        show_locals: If True, local variables will be shown in tracebacks.
    """
    rich.pretty.install(console=logging_console())
    rich.traceback.install(console=logging_console(), show_locals=show_locals)



from ._atexit import register_timer
from ._base import TimerAttrs, TimerTrait
from ._function import TimedFunction
from ._iterable import TimedIterable
from ._records import TimerRecordsAttrs, TimerRecordsTrait
from ._time import TimeCounterName, get_time
from ._timer import timer

__all__ = [
    "TimeCounterName",
    "TimedFunction",
    "TimedIterable",
    "TimerAttrs",
    "TimerRecordsAttrs",
    "TimerRecordsTrait",
    "TimerTrait",
    "get_time",
    "register_timer",
    "timer",
]



import os
from typing import Any

import pydantic

from liblaf import grapes


def load_pydantic[C: pydantic.BaseModel](
    fpath: str | os.PathLike[str], cls: type[C], *, ext: str | None = None
) -> C:
    """Load and deserialize data from a file, then validate it using a Pydantic model.

    Args:
        fpath: The file path to load the data from.
        cls: The Pydantic model class to validate the data against.
        ext: The file extension to use for deserialization.

    Returns:
        An instance of the Pydantic model class with the validated data.
    """
    data: Any = grapes.deserialize(fpath, ext=ext)
    return cls.model_validate(data)


def save_pydantic(
    fpath: str | os.PathLike[str],
    data: pydantic.BaseModel,
    *,
    ext: str | None = None,
    # pydantic.BaseModel.model_dump(**kwargs)
    context: Any | None = None,
    by_alias: bool = False,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
    exclude_none: bool = False,
    round_trip: bool = False,
    serialize_as_any: bool = False,
) -> None:
    """Save a Pydantic model to a file using the specified serialization options.

    Args:
        fpath: The file path where the data should be saved.
        data: The Pydantic model instance to be serialized and saved.
        ext: The file extension to use.
        context: Additional context for the serialization.
        by_alias: Whether to use the alias names defined in the model.
        exclude_unset: Whether to exclude unset fields from the serialized output.
        exclude_defaults: Whether to exclude fields with default values from the serialized output.
        exclude_none: Whether to exclude fields with None values from the serialized output.
        round_trip: Whether to enable round-trip serialization.
        serialize_as_any: Whether to serialize the model as any type.
    """
    grapes.serialize(
        fpath,
        data.model_dump(
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            serialize_as_any=serialize_as_any,
        ),
        ext=ext,
    )



from pathlib import Path

import git
import git.exc


def root() -> Path:
    repo = git.Repo(search_parent_directories=True)
    return Path(repo.working_dir)


def root_safe() -> Path:
    try:
        return root()
    except git.exc.InvalidGitRepositoryError:
        return Path()



from collections.abc import Iterable, Iterator

from . import TimerAttrs, TimerTrait


class TimedIterable[T](TimerTrait):
    iterable: Iterable[T]
    __timer_attrs: TimerAttrs

    @property
    def _timer_attrs(self) -> TimerAttrs:
        return self.__timer_attrs

    def __init__(self, iterable: Iterable[T], timer: TimerTrait) -> None:
        self.__timer_attrs = timer._timer_attrs  # noqa: SLF001
        self.iterable = iterable
        self.label = self.label or "Iterable"

    def __contains__(self, x: object, /) -> bool:
        return x in self.iterable  # pyright: ignore[reportOperatorIssue]

    def __iter__(self) -> Iterator[T]:
        for item in self.iterable:
            self.start()
            yield item
            self.end(depth=3)
        self.log_summary(depth=2)

    def __len__(self) -> int:
        return len(self.iterable)  # pyright: ignore[reportArgumentType]



import contextlib
import inspect
import itertools
import logging
from collections.abc import Iterable, Sequence

import loguru
from environs import Env
from loguru import logger
from rich.console import Console

from liblaf import grapes

type Filter = "str | loguru.FilterDict | loguru.FilterFunction"


DEFAULT_FILTER: Filter = {
    "": "INFO",
    "__main__": "TRACE",
    "liblaf": "DEBUG",
}


DEFAULT_LEVELS: Sequence["loguru.LevelConfig"] = [
    {"name": "ICECREAM", "no": 15, "color": "<magenta><bold>", "icon": "🍦"}
]


class InterceptHandler(logging.Handler):
    """A logging handler that intercepts log messages and redirects them to Loguru.

    This handler is designed to be compatible with the standard logging framework and allows the use of Loguru for logging while maintaining compatibility with existing logging configurations.
    """

    # [Overview — loguru documentation](https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging)

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a log record.

        This function is called by the logging framework to handle a log record.
        It maps the standard logging levels to Loguru levels and finds the caller
        frame from where the log message originated. Finally, it logs the message
        using Loguru.

        Args:
            record: The log record to be emitted.
        """
        # Get corresponding Loguru level if it exists.
        level: str | int
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


def add_level(
    name: str, no: int, color: str | None = None, icon: str | None = None
) -> None:
    """Add a new logging level to the logger.

    Args:
        name: The name of the new logging level.
        no: The numeric value of the new logging level.
        color: The color associated with the new logging level.
        icon: The icon associated with the new logging level.
    """
    with contextlib.suppress(ValueError):
        logger.level(name, no, color=color, icon=icon)


def setup_loguru_logging_intercept(
    level: int | str = logging.NOTSET, modules: Iterable[str] = ()
) -> None:
    """Sets up logging interception using Loguru.

    This function configures the logging module to use Loguru for handling log messages. It sets the logging level and replaces the handlers for the specified modules with an InterceptHandler that redirects log messages to Loguru.

    Args:
        level: The logging level to set.
        modules: A list of module names whose loggers should be intercepted.
    """
    # [loguru-logging-intercept/loguru_logging_intercept.py at f358b75ef4162ea903bf7a3298c22b1be83110da · MatthewScholefield/loguru-logging-intercept](https://github.com/MatthewScholefield/loguru-logging-intercept/blob/f358b75ef4162ea903bf7a3298c22b1be83110da/loguru_logging_intercept.py#L35C5-L42)
    logging.basicConfig(level=level, handlers=[InterceptHandler()])
    for logger_name in itertools.chain(("",), modules):
        mod_logger: logging.Logger = logging.getLogger(logger_name)
        mod_logger.handlers = [InterceptHandler(level=level)]
        mod_logger.propagate = False


def init_loguru(
    level: int | str = logging.NOTSET,
    filter: Filter | None = None,  # noqa: A002
    handlers: Sequence["loguru.HandlerConfig"] | None = None,
    levels: Sequence["loguru.LevelConfig"] | None = None,
) -> None:
    """Initialize the Loguru logger with specified configurations.

    Args:
        level: The logging level.
        filter: A filter to apply to the logger.
        handlers: A sequence of handler configurations.
        levels: A sequence of level configurations.
    """
    filter = filter or DEFAULT_FILTER  # noqa: A001
    if handlers is None:
        console: Console = grapes.logging_console()
        handlers: list[loguru.HandlerConfig] = [
            {
                "sink": lambda s: console.print(
                    s, end="", no_wrap=True, crop=False, overflow="ignore"
                ),
                "format": "[green]{time:YYYY-MM-DD HH:mm:ss.SSS}[/green] | [logging.level.{level}]{level: <8}[/logging.level.{level}] | [cyan]{name}[/cyan]:[cyan]{function}[/cyan]:[cyan]{line}[/cyan] - {message}",
                "filter": filter,
            }
        ]
        env: Env = grapes.environ.init_env()
        if fpath := env.path("LOGGING_FILE", None):
            handlers.append({"sink": fpath, "filter": filter, "mode": "w"})
        if fpath := env.path("LOGGING_JSONL", None):
            handlers.append(
                {"sink": fpath, "filter": filter, "serialize": True, "mode": "w"}
            )
    logger.configure(handlers=handlers)
    for lvl in levels or DEFAULT_LEVELS:
        add_level(**lvl)
    setup_loguru_logging_intercept(level=level)
